@@ -1,7 +1,6 @@
 # -*- coding: utf-8 -*-
 """Command-line interface and logging configuration."""
 # standard-library imports
-
 from typing import Optional
 
 import typer
@@ -9,12 +8,13 @@ import typer
 from . import VERSION
 from .common import APP
 from .common import STATE
-from .plddt import plddt_truncate
+from .plddt import plddt_stats
 
 # global constants
-unused_commands = (plddt_truncate,)
+unused_commands = (plddt_stats,)
 click_object = typer.main.get_command(APP)  # noqa: F841
 
+
 def version_callback(value: bool) -> None:
     """Print version info."""
     if value:
@@ -42,9 +42,9 @@ def set_global_state(
         STATE["log_level"] = "DEBUG"
     elif quiet:
         STATE["log_level"] = "ERROR"
-    unused_state_str = f"{version}"  # noqa: F841
+    _state_str = f"{version}"  # noqa: F841
 
 
-def main() -> None:
+def cli() -> None:
     """Run the app."""
     APP()
