@@ -1,35 +1,25 @@
 # -*- coding: utf-8 -*-
-"""Find hydrated waters in structure."""
-# standard library imports
+"""Operations on pLDDT statistics."""
 from pathlib import Path
 from typing import List
 from typing import Optional
 from typing import Tuple
 
-
-# 3rd-party imports
+import Bio.PDB
 import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
-import seaborn as sns
+import seaborn as sns  # type: ignore
 from loguru import logger
 from statsdict import Stat
 
-# module imports
 from . import NAME
 from . import VERSION
 from .common import APP
 from .common import STATS
 
 # global constants
-ATOM_REC: str = "ATOM"
-ATOM_START_POS: int = 13
-ATOM_STOP_POS: int = 15
 ATOMS: Tuple[str, ...] = ("CA",)
-B_FACTOR_START: int = 61
-B_FACTOR_STOP: int = 65
-REC_TYPE_START: int = 0
-REC_TYPE_STOP: int = 4
 DEFAULT_MIN_LENGTH = 20
 DEFAULT_MIN_COUNT = 20
 DEFAULT_PLDDT_LOWER_BOUND = 80
@@ -43,7 +33,11 @@ MODULE_NAME = __name__.split(".")[0]
 EMPTY_PATH = Path()
 
 
-def bin_labels(bin_type, lower_bound, upper_bound=DEFAULT_PLDDT_UPPER_BOUND):
+def bin_labels(
+    bin_type: str,
+    lower_bound: int,
+    upper_bound: Optional[int] = DEFAULT_PLDDT_UPPER_BOUND,
+) -> str:
     """Create labels for bins of different quantities."""
     if upper_bound == DEFAULT_PLDDT_UPPER_BOUND:
         upper_label = ""
@@ -53,31 +47,34 @@ def bin_labels(bin_type, lower_bound, upper_bound=DEFAULT_PLDDT_UPPER_BOUND):
 
 
 def extract_b_factors(file_path: Path) -> List[float]:
-    """Return an array of B factors from a PDB file specified by file_path."""
+    """Return an array of B factors from a model file specified by file_path."""
     if not file_path.exists():
-        raise ValueError(f"PDB file {file_path} does not exist")
-    with file_path.open("rU") as f:
-        # parse b_factors out of PDB file
-        b_factor_list = [
-            float(rec[B_FACTOR_START:B_FACTOR_STOP])
-            for rec in f.readlines()
-            if (
-                (len(rec) > B_FACTOR_STOP)
-                and (rec[REC_TYPE_START:REC_TYPE_STOP] == ATOM_REC)
-                and (rec[ATOM_START_POS:ATOM_STOP_POS] in ATOMS)
-            )
-        ]
+        raise ValueError(f"Model file {file_path} does not exist")
+    file_ext = file_path.suffix
+    if file_ext == ".cif":
+        parser = Bio.PDB.MMCIFParser()
+    elif file_ext == ".pdb":
+        parser = Bio.PDB.PDBParser()
+    else:
+        logger.warn(
+            f'Unrecognized file extension "{file_ext}", treating as PDB'
+        )
+        parser = Bio.PDB.PDBParser()
+    structure = parser.get_structure("AF_model", file_path)
+    b_factor_list = [
+        a.bfactor for a in structure.get_atoms() if a.name in ATOMS
+    ]
     return b_factor_list
 
 
 def compute_plddt_stats(
-    file_path,
-    lower_bound=DEFAULT_PLDDT_LOWER_BOUND,
-    min_count=DEFAULT_MIN_COUNT,
-    min_length=DEFAULT_MIN_LENGTH,
-    upper_bound=DEFAULT_PLDDT_UPPER_BOUND,
-):
-    """Compute stats on pLDDTs for a PDB file specified by file_path."""
+    file_path: Path,
+    lower_bound: float = DEFAULT_PLDDT_LOWER_BOUND,
+    min_count: int = DEFAULT_MIN_COUNT,
+    min_length: int = DEFAULT_MIN_LENGTH,
+    upper_bound: int = DEFAULT_PLDDT_UPPER_BOUND,
+) -> Tuple[int, float, float, float, float, float, float, str]:
+    """Compute stats on pLDDTs for a model file specified by file_path."""
     plddts = np.array(extract_b_factors(file_path))
     n_pts = len(plddts)
     mean = np.NAN
@@ -88,13 +85,12 @@ def compute_plddt_stats(
     trunc_frac = np.NAN
     if n_pts >= min_length:
         mean = plddts.mean().round(2)
-        median = np.median(plddts).round(2)
-        obs = plddts[(plddts >= lower_bound) &
-                     (plddts <= upper_bound)]
+        median = np.median(plddts).round(2)  # type: ignore
+        obs = plddts[(plddts >= lower_bound) & (plddts <= upper_bound)]  # type: ignore
         n_trunc_obs = len(obs)
         if len(obs) >= min_count:
             trunc_mean = obs.mean().round(2)
-            trunc_median = np.median(obs).round(2)
+            trunc_median = np.median(obs).round(2)  # type: ignore
             trunc_frac = round(n_trunc_obs / n_pts, 2)
     return (
         n_pts,
@@ -111,19 +107,21 @@ def compute_plddt_stats(
 @APP.command()
 @STATS.auto_save_and_report
 def plddt_stats(
-    pdb_list: Optional[List[Path]],
-    criterion: Optional[float] = DEFAULT_PLDDT_CRITERION,
-    min_length: Optional[int] = DEFAULT_MIN_LENGTH,
-    min_count: Optional[int] = DEFAULT_MIN_COUNT,
-    lower_bound: Optional[int] = DEFAULT_PLDDT_LOWER_BOUND,
-    upper_bound: Optional[int] = DEFAULT_PLDDT_UPPER_BOUND,
-    file_stem: Optional[str] = MODULE_NAME,
+    model_file_list: List[Path],
+    criterion: float = DEFAULT_PLDDT_CRITERION,
+    min_length: int = DEFAULT_MIN_LENGTH,
+    min_count: int = DEFAULT_MIN_COUNT,
+    lower_bound: int = DEFAULT_PLDDT_LOWER_BOUND,
+    upper_bound: int = DEFAULT_PLDDT_UPPER_BOUND,
+    file_stem: str = MODULE_NAME,
 ) -> None:
-    """Calculate stats on bounded pLDDTs from list of PDB model files."""
+    """Calculate stats on bounded pLDDTs from list of AlphaFold model files."""
     results = []
-    criterion_label = bin_labels(CRITERION_TYPE, lower_bound, upper_bound)
+    criterion_label = bin_labels(
+        CRITERION_TYPE, lower_bound, upper_bound=upper_bound
+    )
     stats_file_path = Path(f"{file_stem}_plddt_stats.tsv")
-    n_models_in = len(pdb_list)
+    n_models_in = len(model_file_list)
     STATS["models_in"] = Stat(n_models_in, desc="models read in")
     STATS["min_length"] = Stat(min_length, desc="minimum sequence length")
     STATS["min_count"] = Stat(
@@ -138,7 +136,7 @@ def plddt_stats(
     STATS["plddt_criterion"] = Stat(
         criterion, desc=f"minimum bounded {CRITERION_TYPE} for selection"
     )
-    for file_path in pdb_list:
+    for file_path in model_file_list:
         results.append(
             compute_plddt_stats(
                 file_path,
@@ -166,12 +164,13 @@ def plddt_stats(
     logger.info(f"Writing stats to {stats_file_path}")
     stats.sort_values(by=criterion_label, inplace=True, ascending=False)
     stats = stats.reset_index()
-    stats.index.name = f"{NAME}-{VERSION}"
-    del stats["index"]
-    if ((lower_bound == DEFAULT_PLDDT_LOWER_BOUND) and
-            (upper_bound == DEFAULT_PLDDT_UPPER_BOUND)):
+    stats.index.name = f"{NAME}-{VERSION}"  # type: ignore
+    del stats["index"]  # type: ignore
+    if (lower_bound == DEFAULT_PLDDT_LOWER_BOUND) and (
+        upper_bound == DEFAULT_PLDDT_UPPER_BOUND
+    ):
         file_col = stats["file"]
-        del stats["file"]
+        del stats["file"]  # type: ignore
         stats["LDDT_expect"] = (
             1.0
             - (
@@ -179,17 +178,17 @@ def plddt_stats(
                 * (1.0 - DEFAULT_LDDT_CRITERION)
                 / (1.0 - DEFAULT_PLDDT_CRITERION / 100.0)
             )
-        ).round(3)
-        stats["passing"] = (stats["LDDT_expect"] >= DEFAULT_LDDT_CRITERION)
+        ).round(  # type: ignore
+            3
+        )
+        stats["passing"] = stats["LDDT_expect"] >= DEFAULT_LDDT_CRITERION
         stats["file"] = file_col
     stats.to_csv(stats_file_path, sep="\t")
     total_residues = int(stats["residues_in_pLDDT"].sum())
     STATS["total_residues"] = Stat(
         total_residues, desc="number of residues in all models"
     )
-    selected_stats = stats[
-        stats[criterion_label] >= criterion
-    ]
+    selected_stats = stats[stats[criterion_label] >= criterion]
     n_models_selected = len(selected_stats)
     frac_models_selected = round(n_models_selected * 100.0 / n_models_in, 0)
     STATS["models_selected"] = Stat(
@@ -208,12 +207,12 @@ def plddt_stats(
 @APP.command()
 @STATS.auto_save_and_report
 def plddt_select_residues(
-    criterion: Optional[float] = DEFAULT_PLDDT_CRITERION,
-    min_length: Optional[int] = DEFAULT_MIN_LENGTH,
-    min_count: Optional[int] = DEFAULT_MIN_COUNT,
-    lower_bound: Optional[int] = DEFAULT_PLDDT_LOWER_BOUND,
-    upper_bound: Optional[int] = DEFAULT_PLDDT_UPPER_BOUND,
-    file_stem: Optional[str] = MODULE_NAME,
+    criterion: float = DEFAULT_PLDDT_CRITERION,
+    min_length: int = DEFAULT_MIN_LENGTH,
+    min_count: int = DEFAULT_MIN_COUNT,
+    lower_bound: int = DEFAULT_PLDDT_LOWER_BOUND,
+    upper_bound: int = DEFAULT_PLDDT_UPPER_BOUND,
+    file_stem: str = MODULE_NAME,
 ) -> None:
     """Select residues from files matching criterion."""
     stats_file_path = Path(f"{file_stem}_plddt_stats.tsv")
@@ -223,7 +222,7 @@ def plddt_select_residues(
     plddt_list = []
     residue_list = []
     file_list = []
-    for row_num, row in stats.iterrows():
+    for _row_num, row in stats.iterrows():
         if (
             (row["residues_in_pLDDT"] >= min_length)
             and (row[criterion_label] >= criterion)
@@ -240,17 +239,21 @@ def plddt_select_residues(
     out_file_path = Path(f"{file_stem}_plddt{lower_bound}_{criterion}.tsv")
     logger.info(f"Writing residue file {out_file_path}")
     df.to_csv(out_file_path, sep="\t")
+    n_select_residues = len(df[df["pLDDT"] >= criterion])
+    per_residue_val = 100 - int(round(n_select_residues * 100.0 / len(df), 0))
+    STATS["usable_residues_pct"] = Stat(
+        100 - per_residue_val, desc=f"residues with LDDT > {criterion}"
+    )
 
 
 @APP.command()
-@STATS.auto_save_and_report
 def plddt_plot_dists(
-    criterion: Optional[float] = DEFAULT_PLDDT_CRITERION,
-    lower_bound: Optional[int] = DEFAULT_PLDDT_LOWER_BOUND,
-    upper_bound: Optional[int] = DEFAULT_PLDDT_UPPER_BOUND,
-    file_stem: Optional[str] = MODULE_NAME,
-    out_file_type: Optional[str] = DEFAULT_OUT_FILE_TYPE,
-    residue_criterion: Optional[int] = DEFAULT_RESIDUE_CRITERION
+    criterion: float = DEFAULT_PLDDT_CRITERION,
+    lower_bound: int = DEFAULT_PLDDT_LOWER_BOUND,
+    upper_bound: int = DEFAULT_PLDDT_UPPER_BOUND,
+    file_stem: str = MODULE_NAME,
+    out_file_type: str = DEFAULT_OUT_FILE_TYPE,
+    residue_criterion: int = DEFAULT_RESIDUE_CRITERION,
 ) -> None:
     """Plot histograms of per-model and per-residue pLDDT distributions."""
     stats_file_path = Path(f"{file_stem}_plddt_stats.tsv")
@@ -265,19 +268,18 @@ def plddt_plot_dists(
     select_residues = pd.read_csv(res_file_path, sep="\t")
     n_select = len(per_model[per_model[criterion_label] >= criterion])
     n_models = len(per_model)
-    n_select_residues = len(select_residues[select_residues["pLDDT"] >= residue_criterion])
+    n_select_residues = len(
+        select_residues[select_residues["pLDDT"] >= residue_criterion]
+    )
     fig = plt.figure()
     ax = fig.add_subplot(111)
-    sns.ecdfplot(data=per_model,
-                 x=x_axis,
-                 ax=ax,
-                 color="darkblue")
+    sns.ecdfplot(data=per_model, x=x_axis, ax=ax, color="darkblue")
     sns.ecdfplot(
         data=select_residues,
         x="pLDDT",
         ax=ax,
         linestyle="dashed",
-        color="orange"
+        color="orange",
     )
     if upper_bound != DEFAULT_PLDDT_UPPER_BOUND:
         upper_bound_str = f"-{upper_bound}"
@@ -285,35 +287,43 @@ def plddt_plot_dists(
         upper_bound_str = ""
     ax.legend(
         labels=[
-            rf'$pLDDT_{{{lower_bound}{upper_bound_str}}}$ of {n_models} "{file_stem}" models',
+            (
+                rf"$pLDDT_{{{lower_bound}{upper_bound_str}}}$ of {n_models}"
+                + rf' "{file_stem}" models'
+            ),
             rf"$pLDDT$ by residue of {n_select} passing models",
         ]
     )
-    per_model_val = 100 - round(int(n_select*100./n_models), 0)
-    hoffset = -1
-    voffset = -10
-    ax.vlines(criterion, 0.0, per_model_val/100., color="darkblue")
-    ax.text(criterion-hoffset, (per_model_val - voffset)/200.,
-            (rf"$pLDDT_{{{lower_bound}{upper_bound_str}}}$" +
-             f"\n = {criterion},\n" +
-             f"{100-per_model_val}% pass\n" +
-             "by model"
-             ),
-            color="darkblue")
-    per_residue_val = 100 - int(round(n_select_residues*100./len(select_residues), 0))
-    ax.vlines(residue_criterion,
-              0.0,
-              per_residue_val/100.,
-              color="orange")
-    voffset = 10
-    ax.text(residue_criterion-hoffset,
-            (per_residue_val-voffset)/200.,
-            (rf"$pLDDT$" +
-             f"\n= {residue_criterion},\n" +
-             f"{100-per_residue_val}% pass\n" +
-             "by residue"
-             ),
-            color="orange")
+    per_model_val = 100 - round(int(n_select * 100.0 / n_models), 0)
+    h_offset = -1
+    v_offset = -10
+    ax.vlines(criterion, 0.0, per_model_val / 100.0, color="darkblue")  # type: ignore
+    ax.text(  # type: ignore
+        criterion - h_offset,
+        (per_model_val - v_offset) / 200.0,
+        (
+            rf"$pLDDT_{{{lower_bound}{upper_bound_str}}}$"
+            + f"\n = {criterion},\n"
+            + f"{100-per_model_val}% pass\n"
+            + "by model"
+        ),
+        color="darkblue",
+    )
+    per_residue_val = 100 - int(
+        round(n_select_residues * 100.0 / len(select_residues), 0)
+    )
+    v_offset = 10
+    ax.text(  # type: ignore
+        residue_criterion - h_offset,
+        (per_residue_val - v_offset) / 200.0,
+        (
+            r"$pLDDT$"
+            + f"\n= {residue_criterion},\n"
+            + f"{100-per_residue_val}% pass\n"
+            + "by residue"
+        ),
+        color="orange",
+    )
     logger.info(f"Saving {fig_file_path}")
     sns.despine()
     plt.savefig(fig_file_path, dpi=300)
